/*
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to breakpoint().  Breakpoint()
 *  simulates a breakpoint by executing a trap #1.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 *    BAA..AA       Set a breakpoint at address AA.AA      OK or ENN
 *
 *    bAA..AA       Remove the breakpoint at address AA.AA OK or ENN
 *
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <node.h>
#include <asm.h>
#include <conc.h>
#include "trap.h"

#ifdef DEBUG
#define debug_msg fprintf
#else
debug_msg() {}
#endif

/************************************************************************
 *
 * external low-level support routines
 */

extern putDebugChar();   /* write a single character      */
extern getDebugChar();   /* read and return a single char */

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
#define BUFMAX 2048

static int initialized = 0;	/* !0 means we've been initialized */

static const char hexchars[]="0123456789abcdef";

#define NUMREGS 10

/* Number of bytes of registers.  */
#define NUMREGBYTES ((NUMREGS+3) * 4)
enum regnames { IPTR, WPTR, AREG, BREG, CREG, STATUS, FPSTATUS, FAREG, FBREG, FCREG };

#define BREAK_INSTN		0
#define BRKMAX	10
#define SINGLE_STEP_ON	(1<<26)
#define SINGLE_STEP_OFF ~(SINGLE_STEP_ON)
struct {
	int InUse;
	unsigned long Address;
	unsigned char Contents;
}BreakStruct[BRKMAX];
#define J_ZERO	0


static unsigned char TrapHandlerStack[TRAP_HANDLER_STACKSIZE];

#define t_Break	0x2
#define t_Error	0x0
#define t_Step	0x10

#define etIntegerError		0x5
#define etIntegerOverFlow	0x6
#define etFPError		0x7
#define etFPInvalidOp		0x8
#define etFPDivideByZero	0x9
#define etFPOverflow		0xa
#define etFPInexact		0xc
#define etFPUnderFlow		0xb
#define etIllegalInstruction	0x2
#define etUnalign		0x3
#define etMemSemError		0xd

/* Convert ch from a hex digit to an int */

static int
hex(ch)
     unsigned char ch;
{
  if (ch >= 'a' && ch <= 'f')
    return ch-'a'+10;
  if (ch >= '0' && ch <= '9')
    return ch-'0';
  if (ch >= 'A' && ch <= 'F')
    return ch-'A'+10;
  return -1;
}

#ifdef GDB_SLOW
getpacket(buffer)
     char *buffer;
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int i;
  int count;
  unsigned char ch;

  do
    {
      /* wait around for the start character, ignore all other characters */
      while ((ch = getDebugChar()) != '$') ;

      checksum = 0;
      xmitcsum = -1;

      count = 0;

      /* now, read until a # or end of buffer is found */
      while (count < BUFMAX)
	{
	  ch = getDebugChar();
	  if (ch == '#')
	    break;
	  checksum = checksum + ch;
	  buffer[count] = ch;
	  count = count + 1;
	}

      if (count >= BUFMAX)
	continue;

      buffer[count] = 0;

      if (ch == '#')
	{
	  xmitcsum = hex(getDebugChar()) << 4;
	  xmitcsum |= hex(getDebugChar());
	  if (checksum != xmitcsum)
	    putDebugChar('-');	/* failed checksum */
	  else
	    {
	      putDebugChar('+'); /* successful transfer */
	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  putDebugChar(buffer[0]);
		  putDebugChar(buffer[1]);
		  /* remove sequence chars from buffer */
		  count = strlen(buffer);
		  for (i=3; i <= count; i++)
		    buffer[i-3] = buffer[i];
		}
	    }
	}
    }
  while (checksum != xmitcsum);
}

/* send the packet in buffer.  */

void
putpacket(buffer)
     unsigned char *buffer;
{
  unsigned char checksum;
  int count;
  unsigned char ch;

  /*  $<packet info>#<checksum>. */
  do
    {
      putDebugChar('$');
      checksum = 0;
      count = 0;

      while (ch = buffer[count])
	{
	  if (! putDebugChar(ch))
	    return;
	  checksum += ch;
	  count += 1;
	}

      putDebugChar('#');
      putDebugChar(hexchars[checksum >> 4]);
      putDebugChar(hexchars[checksum & 0xf]);

    }
  while (getDebugChar() != '+');
}
#else
void
getpacket(buffer)
     char *buffer;
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int i;
  int count, len;
  unsigned char ch;
  unsigned char tempBuf[3];

  do
    {
#if 0
      /* wait around for the start character, ignore all other characters */
      while ((ch = getDebugChar()) != '$') ;

      checksum = 0;
      xmitcsum = -1;

      getDebugBuf( tempBuf, 2 );
      count = hex(tempBuf[0]) << 4;
      count |= hex(tempBuf[1]);
      count = getDebugBuf( buffer, count );
#else
      do {
         len = getDebugBuf( tempBuf, 3 );
      } while (( tempBuf[0] != '$' )&&( tempBuf[1] != '$' )&&( tempBuf[2] != '$' ));
      if ( tempBuf[1] == '$' ){
        tempBuf[1] = tempBuf[2];
        tempBuf[2] = getDebugChar();
      }
      else if ( tempBuf[2] == '$' ){
	      getDebugBuf( tempBuf+1, 2 );
      }
      count = hex(tempBuf[1]) << 4;
      count |= hex(tempBuf[2]);
      len = count+3;	/* <packet>#<checksum> */
      checksum = 0;
      xmitcsum = -1;
      len = getDebugBuf( buffer, len );
#endif

      /* now, read until a # or end of buffer is found */
      for ( i = 0; i < len; i++ )
	{
          ch = buffer[i];
	  if (ch == '#')
	    break;
	  checksum = checksum + ch;
	}

      if (i > count)
	continue;

      buffer[count] = 0;

      if (ch == '#')
	{
	  xmitcsum = hex(buffer[i+1]) << 4;
	  xmitcsum |= hex(buffer[i+2]);
	  if (checksum != xmitcsum)
	    putDebugChar('-');	/* failed checksum */
	  else
	    {
	      tempBuf[0] = '+'; /* successful transfer */
	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  tempBuf[1] = buffer[0];
		  tempBuf[2] = buffer[1];
                  putDebugBuf( tempBuf, 3 );
		  /* remove sequence chars from buffer */
		  count = strlen(buffer);
		  for (i=3; i <= count; i++)
		    buffer[i-3] = buffer[i];
		}
              else
                putDebugChar( '+' );
	    }
	}
    }
  while (checksum != xmitcsum);
}

/* send the packet in buffer.  */

void
putpacket(buffer)
     unsigned char *buffer;
{
  unsigned char checksum;
  int count, len;
  unsigned char ch;
  unsigned char *tempBuf;

  checksum = 0;
  count = 0;
  while (ch = buffer[count])
	{
	  checksum += ch;
	  count += 1;
	}
  count += 4;
  tempBuf = (unsigned char *)malloc( count * sizeof(unsigned char));
  tempBuf[0] = '$';
  count = 0;
  while (ch = buffer[count])
	{
          tempBuf[count+1] = ch;
	  count += 1;
	}
  count++;
  tempBuf[count++] = '#';
  tempBuf[count++] = hexchars[checksum >> 4];
  tempBuf[count++] = hexchars[checksum & 0xf];
  /*  $<packet info>#<checksum>. */
  do
    {
       putDebugBuf( tempBuf, count );
    }
  while (getDebugChar() != '+');
}
#endif

static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/* Indicate to caller of mem2hex or hex2mem that there has been an
   error.  */
static volatile int mem_err = 0;

/* Convert the memory pointed to by mem into hex, placing result in buf.
 * Return a pointer to the last char put in buf (null), in case of mem fault,
 * return 0.
 * If MAY_FAULT is non-zero, then we will handle memory faults by returning
 * a 0, else treat a fault like any other fault in the stub.
 */

static unsigned char *
mem2hex(mem, buf, count, may_fault)
     unsigned char *mem;
     unsigned char *buf;
     int count;
     int may_fault;
{
  unsigned char ch;

  while (count-- > 0)
    {
      ch = *mem++;
      if (mem_err)
	return 0;
      *buf++ = hexchars[ch >> 4];
      *buf++ = hexchars[ch & 0xf];
    }

  *buf = 0;

  return buf;
}

/* convert the hex array pointed to by buf into binary to be placed in mem
 * return a pointer to the character AFTER the last byte written */

static char *
hex2mem(buf, mem, count, may_fault)
     unsigned char *buf;
     unsigned char *mem;
     int count;
     int may_fault;
{
  int i;
  unsigned char ch;

  for (i=0; i<count; i++)
    {
      ch = hex(*buf++) << 4;
      ch |= hex(*buf++);
      *mem++ = ch;
      if (mem_err)
	return 0;
    }

  return mem;
}


/* Set up exception handlers for tracing and breakpoints */

extern void DebugTrapHandler();
#if 0
extern void DefaultTrapHandler();

void SetupTrapHandler()
{
	THDS *th;
	int Wptr;

	Wptr = (int)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-64-sizeof(THDS)]);
	GetMDL(Wptr+4);
	th = (THDS *)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-sizeof(THDS)]);
	th->Fptr = th->Bptr = NOT_PROCESS_P;
	th->Iptr = (unsigned long)DefaultTrapHandler;
	th->eWu = th->eWl = 0;
	th->Cntl = IGNORE_ALL_TRAPS;
	SetTrapHandler( th );
	th = ((THDS * (*) (const char *)) __asm)( "ldth" );
}
#endif


void
set_debug_traps()
{
	THDS *th;
	int Wptr;

	Wptr = (int)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-64-sizeof(THDS)]);
	GetMDL(Wptr+4);
	th = (THDS *)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-sizeof(THDS)]);
	th->Fptr = th->Bptr = NOT_PROCESS_P;
	th->Iptr = (unsigned long)DebugTrapHandler;
	th->eWu = th->eWl = 0;
	th->Cntl = REPORT_ALL_TRAPS;
	SetTrapHandler( th );


        debug_msg( stderr, "th %x\n", th );
	th = ((THDS * (*) (const char *)) __asm)( "ldth" );
	debug_msg( stderr, "th %x trap handler set\n", th );

	initialized = 1;
}



static int
computeSignal(TrapReason, ErrorType)
     unsigned long TrapReason, ErrorType;
{
  if (( TrapReason & t_Break )||( TrapReason & t_Step ))
	  return SIGTRAP;
  if (( ErrorType == etIntegerError )||
      ( ErrorType == etIntegerOverFlow )||
      ( ErrorType == etFPError )||
      ( ErrorType == etFPInvalidOp )||
      ( ErrorType == etFPDivideByZero )||
      ( ErrorType == etFPOverflow )||
      ( ErrorType == etFPInexact )||
      ( ErrorType == etFPUnderFlow ))
	return SIGFPE;
  else if ( ErrorType == etIllegalInstruction )
	return SIGILL;
  else if ( ErrorType == etUnalign)
	return SIGSEGV;
  else
	return  SIGTRAP;
      
}

/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */

static int
hexToInt(char **ptr, int *intValue)
{
  int numChars = 0;
  int hexValue;

  *intValue = 0;

  while (**ptr)
    {
      hexValue = hex(**ptr);
      if (hexValue < 0)
	break;

      *intValue = (*intValue << 4) | hexValue;
      numChars ++;

      (*ptr)++;
    }

  return (numChars);
}

/*
 * This function does all command procesing for interfacing to gdb.  It
 * returns 1 if you should skip the instruction at the trap address, 0
 * otherwise.
 */

extern void breakinst();
static int first_time = 1;

void
handle_exception (th, TrapReason, ErrorType)
THDS *th;
unsigned long TrapReason;
unsigned long ErrorType;
{
  int sigval;
  int addr;
  int length;
  char *ptr;
  unsigned long *sp;
  int i;
  static int Initialized = 0;
  unsigned long iptr;
  char regs_modified = 0;
  int ThreadId = 0;
  unsigned long floatreg[7];

  SaveFloatReg( floatreg );
  TrapReason &= 0x1f;
  iptr = th->Eptr;
  debug_msg( stderr, "THDS %x TrapReason %x\n", th, TrapReason );
  debug_msg( stderr, "Trap Iptr %x Eptr %x Wptr %x\n", th->sIptr, th->Eptr, th->sWptr );
  /* reply to host that an exception has occurred */
  sigval = computeSignal(TrapReason, ErrorType);

  if ( !Initialized ){
    ptr = remcomOutBuffer;
    *ptr++ = 'N';
    *ptr++ = hexchars[sigval >> 4];
    *ptr++ = hexchars[sigval & 0xf];
    sprintf (remcomOutBuffer + 3, "0x%x;0x%x;0x%x",
               _Node.process->code,
               _Node.process->data,
               _Node.process->data);
    putpacket(remcomOutBuffer);
    Initialized = 1;
    debug_msg( stderr, "Relocate %s\n", remcomOutBuffer + 3 );
    debug_msg( stderr, "Relocate %s, datasize %x\n", remcomOutBuffer+3,  _Node.process->datasize );
  }

  ptr = remcomOutBuffer;
  *ptr++ = 'T';
  *ptr++ = hexchars[sigval >> 4];
  *ptr++ = hexchars[sigval & 0xf];

  *ptr++ = hexchars[ThreadId >> 4];
  *ptr++ = hexchars[ThreadId & 0xf];
  *ptr++ = ';';

  *ptr++ = hexchars[IPTR >> 4];
  *ptr++ = hexchars[IPTR & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->Eptr), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[WPTR >> 4];
  *ptr++ = hexchars[WPTR & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->sWptr), ptr, 4, 0); /* WPTR */
  *ptr++ = ';';

  *ptr++ = hexchars[AREG >> 4];
  *ptr++ = hexchars[AREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->sAreg), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[BREG >> 4];
  *ptr++ = hexchars[BREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->sBreg), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[CREG >> 4];
  *ptr++ = hexchars[CREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->sCreg), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[STATUS >> 4];
  *ptr++ = hexchars[STATUS & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&(th->Cntl), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[FPSTATUS >> 4];
  *ptr++ = hexchars[FPSTATUS & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)(floatreg), ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[FAREG >> 4];
  *ptr++ = hexchars[FAREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)(floatreg+1), ptr, 8, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[FBREG >> 4];
  *ptr++ = hexchars[FBREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)(floatreg+3), ptr, 8, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[FCREG >> 4];
  *ptr++ = hexchars[FCREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)(floatreg+5), ptr, 8, 0);
  *ptr++ = ';';


  *ptr++ = 0;

  putpacket(remcomOutBuffer);

  while (1)
    {
      remcomOutBuffer[0] = 0;

      getpacket(remcomInBuffer);
      debug_msg( stderr, "Received %s\n", remcomInBuffer );
      switch (remcomInBuffer[0])
	{
	case '?':
	  ptr = remcomOutBuffer;
	  *ptr++ = 'N';
	  *ptr++ = hexchars[sigval >> 4];
	  *ptr++ = hexchars[sigval & 0xf];
	  sprintf (remcomOutBuffer + 3, "0x%x;0x%x;0x%x",
		       _Node.process->code,
		       _Node.process->data,
		       _Node.process->data);
	  break;

	case 'd':
				/* toggle debug flag */
	  break;

	case 'g':		/* return the value of the CPU registers */
	  {
	    ptr = remcomOutBuffer;
	    ptr = mem2hex((char *)&(iptr), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(th->sWptr), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(th->sAreg), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(th->sBreg), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(th->sCreg), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(th->Cntl), ptr, 4, 0); 
	    ptr = mem2hex((char *)&(floatreg), ptr, 4*7, 0); 
	  }
	  break;

	case 'G':	   /* set the value of the CPU registers - return OK */
	  {
	    unsigned long *newsp, psr;

	    sp = th->sWptr;
	    regs_modified = 1;

	    ptr = &remcomInBuffer[1];
	    hex2mem(ptr, (char *)&(iptr), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(th->sWptr), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(th->sAreg), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(th->sBreg), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(th->sCreg), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(th->Cntl), 4, 0);  ptr += 8;
	    hex2mem(ptr, (char *)&(floatreg), 4*7, 0);  ptr += 8;

	    /* See if the stack pointer has moved.  If so, then copy the saved
	       locals and ins to the new location.  This keeps the window
	       overflow and underflow routines happy.  */

	    newsp = (unsigned long *)&(th->sWptr);
	    if (sp != newsp)
#if 0
	      sp = memcpy(newsp, sp, 16 * 4);
#else
            {
	      strcpy(remcomOutBuffer,"OK");
              break;
            }
#endif


	    strcpy(remcomOutBuffer,"OK");
	  }
	  break;

	case 'm':	  /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	  /* Try to read %x,%x.  */

	  ptr = &remcomInBuffer[1];

	  if (hexToInt(&ptr, &addr)
	      && *ptr++ == ','
	      && hexToInt(&ptr, &length))
	    {
              if (( addr >= MOST_NEG_INT )&&( addr <= (MOST_NEG_INT + _Node.process->processor->memsize )))
              {
	        if (mem2hex((char *)addr, remcomOutBuffer, length, 1))
		  break;
              }

	      strcpy (remcomOutBuffer, "E03");
	    }
	  else
	    strcpy(remcomOutBuffer,"E01");
	  break;

	case 'M': /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
	  /* Try to read '%x,%x:'.  */

	  ptr = &remcomInBuffer[1];

	  if (hexToInt(&ptr, &addr)
	      && *ptr++ == ','
	      && hexToInt(&ptr, &length)
	      && *ptr++ == ':')
	    {
	      if (hex2mem(ptr, (char *)addr, length, 1))
		strcpy(remcomOutBuffer, "OK");
	      else
		strcpy(remcomOutBuffer, "E03");
	    }
	  else
	    strcpy(remcomOutBuffer, "E02");
	  break;

	case 'c':    
	case 's':    
          /* cAA..AA    Continue at address AA..AA(optional) */
          /* sAA..AA   Step one instruction from AA..AA(optional) */
	  /* try to read optional parameter, pc unchanged if no parm */

	  ptr = &remcomInBuffer[1];
	  if (hexToInt(&ptr, &addr))
	    {
	      th->sIptr = addr;
	    }
          else if ( regs_modified )
	      th->sIptr = iptr;
	  else if (( TrapReason != t_Error )&& (*(char *)th->Eptr != BREAK_INSTN ))
		th->sIptr = th->Eptr;
	  first_time = 0;
          /* clear the trace bit */
          th->Cntl &= SINGLE_STEP_OFF;

          /* set the trace bit if we're stepping */
          if (remcomInBuffer[0] == 's') th->Cntl |= SINGLE_STEP_ON;

	  RestoreFloatReg(floatreg);
	  return;

	  /* kill the program */
	case 'k' :
	  break;
	case 'B' :
	  ptr = &remcomInBuffer[1];
	  hexToInt(&ptr, &addr);

	  for ( i = 0; i < BRKMAX; i++ ){
		if (!BreakStruct[i].InUse ){
		   BreakStruct[i].Address = addr;
		   BreakStruct[i].Contents = *(unsigned char *)addr;
		   BreakStruct[i].InUse = 1;
		   *(unsigned char *)addr = J_ZERO;
	        }
	  }
	  strcpy(remcomOutBuffer, "OK");
	  break;
	case 'b' :
	  ptr = &remcomInBuffer[1];
	  hexToInt(&ptr, &addr);
	  for ( i = 0; i < BRKMAX; i++ ){
		if (BreakStruct[i].InUse ){
		   if ( BreakStruct[i].Address == addr ){
		       *(unsigned char *)addr = BreakStruct[i].Contents;
		       BreakStruct[i].InUse = 0;
		       break;
		   }
	        }
	  }
	  if ( i < BRKMAX )
	    strcpy(remcomOutBuffer, "OK");
	  else
	    strcpy(remcomOutBuffer, "E04");
	  break;

	}			/* switch */

      /* reply to the request */
      putpacket(remcomOutBuffer);
    }
}

/* This function will generate a breakpoint exception.  It is used at the
   beginning of a program to sync up with a debugger and can be used
   otherwise as a quick means to stop program execution and "break" into
   the debugger. */

void
breakpoint()
{
  int temp;
  if (!initialized)
    return;

  /* In case GDB is started before us, ack any packets (presumably
	"$?#xx") sitting there.  */

   putDebugChar ('+');


  __asm ( "j 0" );
}

exceptionHandler()
{
}

