/*{{{  include files*/

#include <conc.h>
#include <string.h>
#include <stdlib.h>
#include "vcpdefs.h"
#include "queue.h"
#include "vcptype.h"
#include "tclVcpDefs.h"
#include "tclVcpType.h"
#include "tclVcpMac.h"
#include "tclInt.h"
#include "debug.h"
/*}}}*/

/*{{{  globals*/
/*
 --   ---------------------------------------------------------------------------
 --
 --      link.c   -   INMOS standard link interface
 --
 --   ---------------------------------------------------------------------------
%W% %G%
*/
/*	Number of ticks in one-tenth of a second */
#define NUM_TICKS		(100*TICKS_PER_MILLISEC)
#define VCP_LINK_NUMBER		4
#define TRANSACTION_BUFFER_SIZE		1024


typedef struct server_link_info {
	QUEUE_CHAIN 	qlink;
	int ServerId;
	HEADER SendHeader, RecvAckHeader;
	VIRTUAL_CHAN SendAckHeader, RecvHeader;
	int SendLink, RecvLink;
}SERVER_LINK_INFO;


static int NextServerId = VCP_LINK_NUMBER;
static SERVER_LINK_INFO *ServerLinkInfo = NULL;

static Channel *InChannel[] = { LINK0IN, LINK1IN, LINK2IN, LINK3IN };
static Channel *OutChannel[] = { LINK0OUT, LINK1OUT, LINK2OUT, LINK3OUT };

/*}}}*/

/*{{{  utilities*/

/*	GET_SERVER_LINK_INFO :
 *		Get the SERVER_LINK_INFO structure corresponding to the
 *		specified LinkId
 */
#define GET_SERVER_LINK_INFO( LinkId, ServerInfo ) {		\
	ServerInfo = ServerLinkInfo;				\
	while ( ServerInfo != NULL ){				\
		if ( ServerInfo->ServerId == LinkId )		\
			break;					\
		else						\
			ServerInfo = QNext(ServerInfo );	\
	}							\
}

/*	GET_ARG :
 *		Get the next argument from the string LinkName,, by
 *	inserting a '\0' at the end of the argument. 
 *	The length of the argument is returned in ArgLen.
 *	NOTE: LinkName is altered.
 */
#define GET_ARG( LinkName, ArgLen ){					\
	ArgLen = 0;							\
	while (( LinkName[ArgLen] != ' ' )&&( LinkName[ArgLen] != '\t' )&&\
	       ( !END_OF_ARG((LinkName+ArgLen))))			\
		ArgLen++;						\
	LinkName[ArgLen] = '\0';					\
	ArgLen++;							\
}

/*
 *	SetServerLink :
 *		Form a server link structure for the link specifications
 *	specified in LinkName. Returns a link identifier. The identifier is
 *	< 4 for raw links, and >=4 for virtual channels.
 */
int SetServerLink( char *LinkName )
{
	int ArgLen;
	SERVER_LINK_INFO *ServerInfo;

	if ( strlen(LinkName) == 2 ){
		/* iserver waiting on raw link */
		return( LinkName[1] - '0' );
	}
	ServerInfo = (SERVER_LINK_INFO *)vcp_malloc( sizeof(SERVER_LINK_INFO) );
	if ( ServerInfo == NULL )
		return( OutOfMemory );
	/* iserver waiting on a virtual channel */
	/* Parse virtual channel info */
	GET_ARG( LinkName, ArgLen );
	GET_LINK_NUMBER( LinkName, ServerInfo->SendLink );
	FORM_HEADER( LinkName, ServerInfo->SendHeader );
	LinkName += ArgLen;
	GET_ARG( LinkName, ArgLen );
	FORM_VIRTUAL_CHAN_ADDR( LinkName, ServerInfo->SendAckHeader );
	LinkName += ArgLen;
	GET_ARG( LinkName, ArgLen );
	GET_LINK_NUMBER( LinkName, ServerInfo->RecvLink );
	FORM_VIRTUAL_CHAN_ADDR( LinkName, ServerInfo->RecvHeader );
	LinkName += ArgLen;
	GET_ARG( LinkName, ArgLen );
	FORM_HEADER( LinkName, ServerInfo->RecvAckHeader );
	ServerInfo->ServerId = NextServerId++;
	EnQueue( ServerLinkInfo, ServerInfo )
	return( ServerInfo->ServerId );
}
/*}}}*/

/*{{{  OpenLink*/
/*
 *
 *   OpenLink
 *
 *   Ready the link associated with `Name'.
 *   If `Name' is NULL or "" then any free link can be used.
 *   Returns any positive integer as a link id or
 *   a negative value if the open fails.
 *
 */
int OpenLink ( Name )
   char *Name;
{
	int ServerId;

	ServerId = SetServerLink( Name );
        return( ServerId );

}
/*}}}*/
/*{{{  CloseLink*/
/*
 *
 *   CloseLink
 *
 *   Close the active link `LinkId'.
 *   Returns 1 on success or negative if the close failed.
 *
 */
int CloseLink ( LinkId )
   int LinkId;
{
	SERVER_LINK_INFO *ServerInfo;

	if ( LinkId < VCP_LINK_NUMBER )
		return(1);
	GET_SERVER_LINK_INFO( LinkId, ServerInfo );
	if ( ServerInfo != NULL ){
		vcp_free( ServerInfo->SendHeader.HeaderPtr );
		vcp_free( ServerInfo->RecvAckHeader.HeaderPtr );
		QDelete( ServerLinkInfo, ServerInfo );
		return( 1 ) ;
	}
	else
		return(-7);

}
/*}}}*/
/*{{{  ReadLink*/
/*
 *
 *   ReadLink
 *
 *   Read `Count' chars into `Buffer' from the specified link.
 *   LinkId is a vaild link identifier, opened with OpenLink.
 *   `Timeout' is a non negative integer representing tenths
 *   of a second.  A `Timeout' of zero is an infinite timeout.
 *   The timeout is for the complete operation.
 *   If `Timeout' is positive then ReadLink may return having
 *   read less that the number of chars asked for.
 *   Returns the number of chars placed in `Buffer' (which may
 *   be zero) or negative to indicate an error.
 *
 */
 
int ReadLink ( LinkId, Buffer, Count, Timeout )
   int LinkId;
   char *Buffer;
   unsigned int Count;
   int Timeout;
{
	int ret;
	Boolean EOMFlag;
	int ReadCount;
	SERVER_LINK_INFO *ServerInfo;

	if ( LinkId < VCP_LINK_NUMBER ){
		if ( Timeout > 0 ){
			ret = ChanInTimeFail( InChannel[LinkId], Buffer, Count, Timeout*NUM_TICKS );
			if ( ret >= 0 )
				return( Count );
			else 
				return( VCP_TimeOut );
		}
		else {
			ChanIn( InChannel[LinkId], Buffer, Count );
			return(Count);
		}
	}
	else {
		GET_SERVER_LINK_INFO( LinkId, ServerInfo );
		if (ServerInfo == NULL)
			return(-11);
		/* Since address of count is used for VCPRecv, this should 
		   always be a variable. ReadLink is invoked with Count set to 
		   a constant, by the iserver */
		ReadCount = Count; 
		ret = VCP_ReceiveMessage( ServerInfo->RecvLink , ServerInfo->RecvHeader, 
			&(ServerInfo->RecvAckHeader), Buffer, &ReadCount, &EOMFlag,Timeout*NUM_TICKS );

		if ( ret == VCP_OK )
			return( ReadCount );
		else
			return( -2 );
	}
}
/*}}}*/
/*{{{  WriteLink*/
/*
 *
 *   WriteLink
 *
 *   Write `Count' chars from `Buffer' to the specified link.
 *   LinkId is a vaild link identifier, opened with OpenLink.
 *   `Timeout' is a non negative integer representing tenths
 *   of a second.  A `Timeout' of zero is an infinite timeout.
 *   The timeout is for the complete operation.
 *   If `Timeout' is positive then WriteLink may return having
 *   written less that the number of chars asked for.
 *   Returns the number of chars actually written (which may
 *   be zero) or negative to indicate an error.
 *
 */
 
int WriteLink ( LinkId, Buffer, Count, Timeout )
   int LinkId;
   char *Buffer;
   unsigned int Count;
   int Timeout;
{
	int ret;
	int i;
	SERVER_LINK_INFO *ServerInfo;

	if ( LinkId < VCP_LINK_NUMBER ){
		if (Count < 1)
			return(-2);
		if ( Timeout > 0 ){
			ret = ChanOutTimeFail( OutChannel[LinkId], Buffer, Count, Timeout*NUM_TICKS );
			if ( ret == 0 )
				return( Count );
			else 
				return(VCP_TimeOut);
		}
		else {
			ChanOut( OutChannel[LinkId], Buffer, Count );
			return( Count );
		}
	}
	else {
		GET_SERVER_LINK_INFO( LinkId, ServerInfo );
		if (ServerInfo == NULL)
			return(-11);
		ret = VCP_SendMessage( ServerInfo->SendLink, &(ServerInfo->SendHeader), ServerInfo->SendAckHeader, 
			Buffer, Count, True, Timeout*NUM_TICKS );
		if ( ret == VCP_OK )
			return( Count );
		else
			return( ret );
	}
}
/*}}}*/
/*{{{  ResetLink*/
/*
 *
 *   ResetLink
 *
 *   Reset the subsystem associated with the specified link.
 *   Returns 1 if the reset is successful, negative otherwise.
 *
 */
 
int ResetLink ( LinkId )
   int LinkId;
{
	SERVER_LINK_INFO *ServerInfo;

	if ( LinkId < VCP_LINK_NUMBER ){
		return(1);
	}
	GET_SERVER_LINK_INFO( LinkId, ServerInfo );
        if (ServerInfo == NULL)
                return(-11);
        return( 1 );

}
/*}}}*/
/*{{{  AnalyseLink*/
/*
 *
 *   AnalyseLink
 *
 *   Analyse the subsystem associated with the specified link.
 *   Returns 1 if the analyse is successful, negative otherwise.
 *
 */
 
int AnalyseLink ( LinkId )
   int LinkId;
{
}
/*}}}*/
/*{{{  TestError*/
/*
 *
 *   TestError
 *
 *   Test the error status associated with the specified link.
 *   Returns 1 if error is set, 0 if it is not and
 *   negative to indicate an error.
 *
 */
 
int TestError ( LinkId )
   int LinkId;
{
}
/*}}}*/
/*{{{  TestRead*/
/*
 *
 *   TestRead
 *
 *   Test input status of the link.
 *   Returns 1 if ReadLink will return one byte without timeout,
 *   0 if it may not and negative to indicate an error.
 *
 */
int TestRead ( LinkId )
   int LinkId;
{
        return(1);

}
/*}}}*/
/*{{{  TestWrite*/
/*
 *
 *   TestWrite
 *
 *   Test output status of the link.
 *   Returns 1 if WriteLink can write one byte without timeout,
 *   0 if it may not and negative to indicate an error.
 *
 */
int TestWrite ( LinkId )
   int LinkId;
{
        return(1);

}

/*
 *   Eof
 */
/*}}}*/
